from paco.models import schemas, vocabulary
from paco.models.base import Deployable, Parent, Named, Name, Resource, ApplicationResource, AccountRef, Regionalized
from paco.models.formatter import get_formatted_model_context
from paco.models.logging import CloudWatchLogSets
from paco.models.locations import get_parent_by_interface
from zope.interface import implementer
from zope.schema.fieldproperty import FieldProperty
import json
import paco.models.exceptions
import paco.models.services
import paco.models.registry
import troposphere
import troposphere.cloudwatch


class Monitor(Named, dict):
    pass

@implementer(schemas.ISNSTopics)
class SNSTopics(AccountRef, Named, dict):
    "Container for SNS Topics"
    regions = FieldProperty(schemas.ISNSTopics["regions"])

    def __init__(self, name, parent):
        super().__init__(name, parent)
        self.regions = ['ALL']

    def resolve_ref(self, ref):
        return self.resolve_ref_obj.resolve_ref(ref)

@implementer(schemas.IHealthChecks)
class HealthChecks(Named, dict):
    pass

@implementer(schemas.IAlarmNotifications)
class AlarmNotifications(Named, dict):
    "Container of AlarmNotifications"

@implementer(schemas.IAlarmNotification)
class AlarmNotification(Named):
    "AlarmNotification"
    groups = FieldProperty(schemas.IAlarmNotification["groups"])
    classification = FieldProperty(schemas.IAlarmNotification["classification"])
    severity = FieldProperty(schemas.IAlarmNotification["severity"])

@implementer(schemas.IAlarmSet)
class AlarmSet(Named, dict):
    resource_type = FieldProperty(schemas.IAlarmSet["resource_type"])

    def __init__(self, name, parent):
        super().__init__(name, parent)
        self.notifications = AlarmNotifications('notifications', self)

@implementer(schemas.IAlarmSets)
class AlarmSets(Named, dict):
    "Collection of Alarms"

class AlarmSetsContainer(Named, dict):
    "Collection of AlarmSets"

@implementer(schemas.IAlarm)
class Alarm(Named, Regionalized, Deployable):
    "Alarm"
    classification = FieldProperty(schemas.IAlarm["classification"])
    description = FieldProperty(schemas.IAlarm["description"])
    runbook_url = FieldProperty(schemas.IAlarm["runbook_url"])
    severity = FieldProperty(schemas.IAlarm["severity"])

    def __init__(self, name, parent):
        super().__init__(name, parent)
        self.notifications = AlarmNotifications('notifications', self)
        self.enabled = True

    def _add_notifications_to_groups(self, notifications, groups):
        for notification in notifications.values():
            # do not add if it is filtered out
            if notification.classification and self.classification != notification.classification:
                continue
            if notification.severity and self.severity != notification.severity:
                continue
            for group in notification.groups:
                if group not in groups:
                    groups[group] = None
        return groups

    @property
    def notification_groups(self):
        "A unique list of notification groups that an Alarm is subscribed to"
        groups = {}

        # start with any notifications specific to the alarm
        groups = self._add_notifications_to_groups(self.notifications, groups)

        # add on notifications for the AlarmSet
        # does not exist for certain Alarms (e.g. Route53HealthCheck Alarm)
        alarm_set = get_parent_by_interface(self, schemas.IAlarmSet)
        if alarm_set != None:
            groups = self._add_notifications_to_groups(alarm_set.notifications, groups)

        # For Alarms that belong to an application, check IMonitorConfig and IApplication
        # add on notifications for the Resource
        monitor = get_parent_by_interface(self, schemas.IMonitorConfig)
        if monitor != None:
            groups = self._add_notifications_to_groups(monitor.notifications, groups)

        # add on notifications for the Application
        app = get_parent_by_interface(self, schemas.IApplication)
        if app != None:
            groups = self._add_notifications_to_groups(app.notifications, groups)

        return [key for key in groups.keys()]

    def get_alarm_actions_paco_refs(self, snstopics=None):
        """Return a list of alarm actions in the form of paco.ref SNS Topic ARNs, e.g.

        'paco.ref service.notification.applications.notification.groups.lambda.resources.snstopic.arn'

        This will by default be a list of SNS Topics that the alarm is subscribed to.
        However, if a plugin is registered, it will provide the actions instead.
        """
        # if a Service has registered a custom AlarmActions hook, call that instead
        if paco.models.registry.CW_ALARM_ACTIONS_HOOK != None:
            return paco.models.registry.CW_ALARM_ACTIONS_HOOK(self)

        # default behaviour is to use notification groups directly
        notification_arns = [
            snstopics[group].paco_ref + '.arn' for group in self.notification_groups
        ]
        if len(notification_arns) > 5:
            raise paco.models.exceptions.InvalidPacoProjectFile("""
    Alarm {} has {} actions, but CloudWatch Alarms allow a maximum of 5 actions.

    {}""".format(self.name, len(notification_arns), get_formatted_model_context(self))
            )

        return notification_arns

    @property
    def actions_enabled(self):
        if hasattr(self, 'alarm_actions'):
            if len(self.alarm_actions) > 0:
                return True
        return None

@implementer(schemas.IDimension)
class Dimension(Parent):
    name = FieldProperty(schemas.IDimension["name"])
    value = FieldProperty(schemas.IDimension["value"])

    def __init__(self, __parent__, name='', value=''):
        self.__parent__ = __parent__
        self.name = name
        self.value = value

@implementer(schemas.ISimpleCloudWatchAlarm)
class SimpleCloudWatchAlarm(Parent):
    "CloudWatch Alarm"
    alarm_description = FieldProperty(schemas.ISimpleCloudWatchAlarm["alarm_description"])
    actions_enabled = FieldProperty(schemas.ISimpleCloudWatchAlarm["actions_enabled"])
    metric_name = FieldProperty(schemas.ISimpleCloudWatchAlarm["metric_name"])
    namespace = FieldProperty(schemas.ISimpleCloudWatchAlarm["namespace"])
    period = FieldProperty(schemas.ISimpleCloudWatchAlarm["period"])
    evaluation_periods = FieldProperty(schemas.ISimpleCloudWatchAlarm["evaluation_periods"])
    threshold = FieldProperty(schemas.ISimpleCloudWatchAlarm["threshold"])
    comparison_operator = FieldProperty(schemas.ISimpleCloudWatchAlarm["comparison_operator"])
    statistic = FieldProperty(schemas.ISimpleCloudWatchAlarm["statistic"])
    dimensions = FieldProperty(schemas.ISimpleCloudWatchAlarm["dimensions"])

    def __init__(self, __parent__):
        self.__parent__ = __parent__
        self.dimensions = []


@implementer(schemas.ICloudWatchAlarm)
class CloudWatchAlarm(Alarm):
    "CloudWatch Alarm"
    type = "Alarm"
    comparison_operator = FieldProperty(schemas.ICloudWatchAlarm["comparison_operator"])
    dimensions = FieldProperty(schemas.ICloudWatchAlarm["dimensions"])
    enable_ok_actions = FieldProperty(schemas.ICloudWatchAlarm["enable_ok_actions"])
    enable_insufficient_data_actions = FieldProperty(schemas.ICloudWatchAlarm["enable_insufficient_data_actions"])
    evaluate_low_sample_count_percentile = FieldProperty(schemas.ICloudWatchAlarm["evaluate_low_sample_count_percentile"])
    evaluation_periods = FieldProperty(schemas.ICloudWatchAlarm["evaluation_periods"])
    extended_statistic = FieldProperty(schemas.ICloudWatchAlarm["extended_statistic"])
    namespace = FieldProperty(schemas.ICloudWatchAlarm["namespace"])
    metric_name = FieldProperty(schemas.ICloudWatchAlarm["metric_name"])
    period = FieldProperty(schemas.ICloudWatchAlarm["period"])
    threshold = FieldProperty(schemas.ICloudWatchAlarm["threshold"])
    treat_missing_data = FieldProperty(schemas.ICloudWatchAlarm["treat_missing_data"])
    statistic = FieldProperty(schemas.ICloudWatchAlarm["statistic"])

    troposphere_props = troposphere.cloudwatch.Alarm.props
    cfn_mapping = {
        # 'AlarmName': computed by CloudFormation service so you can specify updates
        # 'AlarmDescription': computed in template for the SNS Topic ARNs
        # 'Dimensions': computed in template,
        # 'AlarmActions': computed in template,
        # 'OKActions': computed in template,
        # 'InsufficientDataActions': copmuted in template,
        'ActionsEnabled': 'actions_enabled',
        'ComparisonOperator': 'comparison_operator',
        'EvaluateLowSampleCountPercentile': 'evaluate_low_sample_count_percentile',
        'EvaluationPeriods': 'evaluation_periods',
        'ExtendedStatistic': 'extended_statistic',
        'MetricName': 'metric_name',
        # 'Namespace': computed in template
        'Period': 'period',
        'Statistic': 'statistic',
        'Threshold': 'threshold',
        'TreatMissingData': 'treat_missing_data',
        # 'Unit': (basestring, False),
        # 'DatapointsToAlarm': (positive_integer, False),
        #  'Metrics': ([MetricDataQuery], False),
    }

    def __init__(self, name, __parent__):
        super().__init__(name, __parent__)
        self._extra_alarm_description = {}

    def threshold_human(self):
        "Human readable threshold"
        comparison = vocabulary.cloudwatch_comparison_operators[self.comparison_operator]
        if self.period < 60:
            period_human = '{} seconds'.format(int(self.period))
        elif self.period < 120:
            period_human = '{} minute'.format(int(self.period / 60))
        elif self.period < 3600:
            period_human = '{} minutes'.format(int(self.period / 60))
        elif self.period < 7200:
            period_human = '{} hour'.format(int(self.period / 3600))
        else:
            period_human = '{} hours'.format(int(self.period / 3600))
        return '{} {} {} for {} datapoints within {}'.format(
            self.metric_name,
            comparison,
            self.threshold,
            self.evaluation_periods,
            period_human
        )

    def get_alarm_description(self, notification_cfn_refs):
        """Create an Alarm Description in JSON format with Paco Alarm information"""
        project = get_parent_by_interface(self, schemas.IProject)
        netenv = get_parent_by_interface(self, schemas.INetworkEnvironment)
        env = get_parent_by_interface(self, schemas.IEnvironment)
        envreg = get_parent_by_interface(self, schemas.IEnvironmentRegion)
        app = get_parent_by_interface(self, schemas.IApplication)
        group = get_parent_by_interface(self, schemas.IResourceGroup)
        resource = get_parent_by_interface(self, schemas.IResource)

        # SNS Topic ARNs are supplied Paramter Refs
        topic_arn_subs = []
        sub_dict = {}
        for action_ref in notification_cfn_refs:
            ref_id = action_ref.data['Ref']
            topic_arn_subs.append('${%s}' % ref_id)
            sub_dict[ref_id] = action_ref

        # Base alarm info - used for standalone alarms not part of an application
        description = {
            "ref": self.paco_ref_parts,
            "project_name": project.name,
            "project_title": project.title,
            "account_name": self.account_name,
            "alarm_name": self.name,
            "classification": self.classification,
            "severity": self.severity,
            "notification_groups": self.notification_groups,
            "topic_arns": topic_arn_subs
        }

        if paco.models.registry.CW_ALARM_DESCRIPTION_HOOK != None:
            description = paco.models.registry.CW_ALARM_DESCRIPTION_HOOK(self, description)

        # conditional fields
        if self.description:
            description['description'] = self.description
        if self.runbook_url:
            description['runbook_url'] = self.runbook_url

        if app != None:
            # Service applications and apps not part of a NetEnv
            description["app_name"] = app.name
            description["app_title"] = app.title
        if group != None:
            # Application level Alarms do not have resource group and resource
            description["resource_group_name"] = group.name
            description["resource_group_title"] = group.title
            description["resource_name"] = resource.name
            description["resource_title"] = resource.title

        if netenv != None:
            # NetEnv information
            description["netenv_name"] = netenv.name
            description["netenv_title"] = netenv.title
            description["env_name"] = env.name
            description["env_title"] = env.title
            description["envreg_name"] = envreg.name
            description["envreg_title"] = envreg.title

        # add any extended fields
        description = self.extend_description(description)

        # bake into JSON
        description_json = json.dumps(description)
        return troposphere.Sub(
            description_json,
            sub_dict
        )

    def add_to_alarm_description(self, add_dict):
        "Adds additional fields to AlarmDescription"
        for key, value in add_dict.items():
            if key in self._extra_alarm_description:
                raise paco.models.exceptions.AlarmDescriptionExtensionConflict(
                   f"Another extension has already added the field name '{key}' to AlarmDescription."
                )
            self._extra_alarm_description[key] = value
        return

    def extend_description(self, description):
        "Extend a description dict with any extra fields"
        for key, value in self._extra_alarm_description.items():
            if key in description:
                raise paco.models.exceptions.AlarmDescriptionExtensionConflict(
                   f"Can not extend AlarmDescription with the field name '{key}' as it's a Paco AlarmDescription reserved name."
                )
            description[key] = value
        return description


@implementer(schemas.ICloudWatchLogAlarm)
class CloudWatchLogAlarm(CloudWatchAlarm):
    log_set_name = FieldProperty(schemas.ICloudWatchLogAlarm["log_set_name"])
    log_group_name = FieldProperty(schemas.ICloudWatchLogAlarm["log_group_name"])

@implementer(schemas.IMonitorable)
class Monitorable():
    monitoring = FieldProperty(schemas.IMonitorable["monitoring"])

@implementer(schemas.IMetric)
class Metric():
    name = FieldProperty(schemas.IMetric["name"])
    measurements = FieldProperty(schemas.IMetric["measurements"])
    collection_interval = FieldProperty(schemas.IMetric["collection_interval"])
    resources = FieldProperty(schemas.IMetric["resources"])
    drop_device = FieldProperty(schemas.IMetric["drop_device"])

    def __init__(self):
        self.resources = []

# AWS Provided Metrics
ec2core_builtin_metric = Metric()
ec2core_builtin_metric.name = 'ec2core_builtin_metric'
ec2core_builtin_metric.collection_interval = 300
ec2core_builtin_metric.measurements = [
    'CPUUtilization',
    'DiskReadBytes',
    'DiskReadOps',
    'DiskWriteBytes',
    'DiskWriteOps',
    'NetworkIn',
    'NetworkOut',
    'StatusCheckFailed',
    'StatusCheckFailed_Instance',
    'StatusCheckFailed_System'
]

asg_builtin_metrics = [
    'GroupMinSize',
    'GroupMaxSize',
    'GroupDesiredCapacity',
    'GroupInServiceInstances',
    'GroupPendingInstances',
    'GroupStandbyInstances',
    'GroupTerminatingInstances',
    'GroupTotalInstances'
]

@implementer(schemas.IMonitorConfig)
class MonitorConfig(Deployable, Named):
    collection_interval = FieldProperty(schemas.IMonitorConfig["collection_interval"])
    metrics = FieldProperty(schemas.IMonitorConfig["metrics"])
    asg_metrics = FieldProperty(schemas.IMonitorConfig["asg_metrics"])
    notifications = FieldProperty(schemas.IMonitorConfig["notifications"])

    def __init__(self, name, __parent__):
        super().__init__(name, __parent__)
        self.alarm_sets = AlarmSets('alarm_sets', self)
        self.health_checks = HealthChecks('health_checks', self)
        self.log_sets = CloudWatchLogSets('log_sets', self)
        self.notifications = AlarmNotifications('notifications', self)
        self.asg_metrics = asg_builtin_metrics
        self.metrics = []

@implementer(schemas.IDashboardVariables)
class DashboardVariables(Named, dict):
    pass

@implementer(schemas.ICloudWatchDashboard)
class CloudWatchDashboard(ApplicationResource):
    dashboard_file = FieldProperty(schemas.ICloudWatchDashboard["dashboard_file"])



"""
Paco References
"""

from paco.models import vocabulary, schemas
from paco.models.locations import get_parent_by_interface
from paco.models.exceptions import InvalidPacoReference
from ruamel.yaml.compat import StringIO
from operator import itemgetter
import importlib
import pathlib
import ruamel.yaml


ami_cache = {}

class YAML(ruamel.yaml.YAML):
    def dump(self, data, stream=None, **kw):
        dumps = False
        if stream is None:
            dumps = True
            stream = StringIO()
        ruamel.yaml.YAML.dump(self, data, stream, **kw)
        if dumps:
            return stream.getvalue()

# ToDo: this is duplicated between reftypes and references
import zope.schema
class InvalidPacoReferenceString(zope.schema.ValidationError):
    __doc__ = 'PacoReference must be of type (string)'

class InvalidPacoReferenceStartsWith(zope.schema.ValidationError):
    __doc__ = "PacoReference must begin with 'paco.ref'"

class InvalidPacoReferenceRefType(zope.schema.ValidationError):
    __doc__ = "PacoReference 'paco.ref must begin with: netenv | resource | accounts | function | service"

def is_ref(paco_ref, raise_enabled=False):
    """Determines if the string value is a Paco reference"""
    if type(paco_ref) != type(str()):
        if raise_enabled: raise InvalidPacoReferenceString
        return False
    if paco_ref.startswith('paco.ref ') == False:
        if raise_enabled: raise InvalidPacoReferenceStartsWith
        return False
    ref_types = ["netenv", "resource", "accounts", "function", "service"]
    for ref_type in ref_types:
        if paco_ref.startswith('paco.ref %s.' % ref_type):
            return True
    if raise_enabled: raise InvalidPacoReferenceRefType
    return False

class Reference():
    """
    Reference to something in the paco.models

    attributes:
      raw : original reference str : 'paco.ref netenv.pacodemo.network.vpc.security_groups.app.lb'
      type : reference type str : 'netenv'
      parts : list of ref parts : ['netenv', 'pacodemo', 'network', 'vpc', 'security_groups', 'app', 'lb']
      ref : reffered string : 'netenv.pacodemo.network.vpc.security_groups.app.lb'
    """

    def __init__(self, value):
        self.raw = value
        self.ref = value.split(' ', 2)[1]
        self.parts = self.ref.split('.')
        self.type = self.parts[0]
        # resource_ref is the tail end of the reference that is
        # relevant to the Resource it references
        self.resource = None
        self.resource_ref = None
        self.region = None

        if self.type == 'netenv' or self.type == 'service':
            # paco.ref service.describe.<account>.<region>.myapp
            # pace.ref netenv.dev.<account>.<region>.applications
            # do not try to find region for short environment refs like 'paco.ref netenv.mynet.prod'
            if len(self.parts) > 3:
                if self.parts[3] in vocabulary.aws_regions.keys():
                    self.region = self.parts[3]
        if is_ref(self.raw) == False:
            print("Invalid Paco reference: %s" % (value))

    def get_account(self, project, resource):
        "Account object this reference belongs to"
        if self.type == 'service':
            return project.accounts[self.parts[2]]
        elif self.type == 'netenv':
            env_reg = get_parent_by_interface(resource, schemas.IEnvironmentRegion)
            return get_model_obj_from_ref(env_reg.network.aws_account, project)
        return None

    @property
    def last_part(self):
        return self.parts[-1]

    def next_part(self, search_ref):
        search_parts = search_ref.split('.')
        first_found = False
        for ref_part in self.parts:
            if first_found == True:
                if search_idx == len(search_parts):
                    return ref_part
                elif ref_part == search_parts[search_idx]:
                    search_idx += 1
            elif ref_part == search_parts[0]:
                first_found = True
                search_idx = 1

        return None

    def sub_part(self, part, value):
        self.raw = self.raw.replace(part, value)
        self.ref = self.ref.replace(part, value)
        self.parts = self.ref.split('.')

    def set_account_name(self, account_name):
        self.sub_part('<account>', account_name)

    def set_environment_name(self, environment_name):
        self.sub_part('<environment>', environment_name)

    def set_region(self, region):
        self.region = region
        self.sub_part('<region>', region)

    def resolve(self, project, account_ctx=None):
        return resolve_ref(
            ref_str=None,
            project=project,
            account_ctx=account_ctx,
            ref=self
        )

    def get_model_obj(self, project):
        return get_model_obj_from_ref(self, project)

    def secret_base_ref(self):
        """
Secret Manager Secrets can be a ref to the Secret:

  paco.ref netenv.mynet.dev.eu-central-1.secrets_manager.myapp.mygroup.mydb

Or they can be a ref to a JSON field within the Secret value:

  paco.ref netenv.mynet.dev.eu-central-1.secrets_manager.myapp.mygroup.mydb.myjsonfield

Either ref type can also specify the ARN of the Secret:

  paco.ref netenv.mynet.dev.eu-central-1.secrets_manager.myapp.mygroup.mydb.arn
  paco.ref netenv.mynet.dev.eu-central-1.secrets_manager.myapp.mygroup.mydb.myjsonfield.arn

Return only the base ref and strip the final part that specifies the JSON Field from the ref.
For all of the above examples, this would return:

  paco.ref netenv.mynet.secrets_manager.myapp.mygroup.mydb

        """
        new_ref_obj = self
        if self.type == 'netenv':
            # remove final .arn but allow for a secret named 'arn'
            #   paco.ref netenv.mynet.dev.eu-central-1.secrets_manager.myapp.mygroup.arn
            #   paco.ref netenv.mynet.dev.eu-central-1.secrets_manager.myapp.mygroup.arn.arn
            #   paco.ref netenv.mynet.dev.eu-central-1.secrets_manager.myapp.mygroup.arn.myjson
            #   paco.ref netenv.mynet.dev.eu-central-1.secrets_manager.myapp.mygroup.arn.myjson.arn
            if len(self.parts) == 7:
                pass
            elif len(self.parts) > 7:
                new_ref = self.parts[:8]
                new_ref = '.'.join(new_ref)
                new_ref_obj = Reference(f'paco.ref {new_ref}')
            else:
                raise InvalidPacoReference("Not a valid Secret ref")
        elif self.type == 'service':
            # service.notification.secrets_manager.myapp.mygroup.mysecret
            if len(self.parts) > 5:
                new_ref = self.parts[:6]
                new_ref = '.'.join(new_ref)
                new_ref_obj = Reference(f'paco.ref {new_ref}')
            else:
                raise InvalidPacoReference("Not a valid Secret ref")
        else:
            raise NotImplemented('What kind of Secret do you have?!?')
        return new_ref_obj

def get_model_obj_from_ref(ref, project, referring_obj=None):
    """Resolves the reference to the model object it refers to.
    ref can be either a string or a Reference object.
    """
    if isinstance(ref, str):
        ref = Reference(ref)
    obj = project
    for part_idx in range(0, len(ref.parts)):
        # special case for resource.route53 hosted zone look-ups
        if schemas.IRoute53Resource.providedBy(obj):
            obj = obj.hosted_zones
        try:
            next_obj = obj[ref.parts[part_idx]]
        except (TypeError, KeyError):
            next_obj = getattr(obj, ref.parts[part_idx], None)
        if next_obj != None and isinstance(next_obj, str) == False:
            obj = next_obj
        else:
            message = f"\nCould not find model at {ref.raw}\n"
            message += f"\nLook-up reached: {obj.paco_ref_parts}\n"
            message += f"\nCould not find next part: {ref.parts[part_idx]}\n"
            if referring_obj != None:
                message += f"Object: {referring_obj.paco_ref_parts}\n"
            if ref.parts[0] in ['iam', 'codecommit', 'ec2', 'sns', 's3', 'route53', 'resources']:
                message += "Did you mean to run:\n"
                message += "paco <command> resource.{}?\n".format(ref.ref)
            raise InvalidPacoReference(message)
    return obj

def get_resolve_ref_obj(project, obj, ref, part_idx_start):
    """
    Traverses the reference parts looking for the last child that
    is a not a string. This object is expected to be a part of the
    model and should have a resolve_ref method that can be called.
    """
    for part_idx in range(part_idx_start, len(ref.parts)):
        # the model can be walked with either obj[name] or obj.name
        # depending on what is being walked
        name = ref.parts[part_idx]
        try:
            next_obj = obj[name]
        except (TypeError, KeyError):
            next_obj = getattr(obj, name, None)
        if next_obj != None and isinstance(next_obj, str) == False and isinstance(next_obj, int) == False:
            obj = next_obj
        else:
            # Given a ref of 'netenv.websites.prod.[...].resources.database.endpoint.address' then
            # after walks to the RDS Resource named '.database' it won't be able to find the next '.endpoint'
            # and will break here
            break

    ref.resource_ref = '.'.join(ref.parts[part_idx:])
    ref.resource = obj
    try:
        response = obj.resolve_ref(ref)
    except AttributeError:
        # Check if we have something stored in Outputs
        outputs_value = resolve_ref_outputs(ref, project['home'])
        if outputs_value != None:
            return outputs_value
        breakpoint()
        raise_invalid_reference(ref, obj, ref.parts[part_idx:][0])
    return response

def resolve_ref_outputs(ref, project_home_path):
    key = ref.parts[0]
    # ToDo: .paco-work is part of paco not paco.models, refactor?
    output_filepath = pathlib.Path(project_home_path) / '.paco-work' / 'outputs' / key
    output_filepath = output_filepath.with_suffix('.yaml')
    if output_filepath.exists() == False:
        return None

    yaml = YAML(typ="safe", pure=True)
    yaml.default_flow_sytle = False
    with open(output_filepath, "r") as output_fd:
        outputs_dict = yaml.load(output_fd)

    for part in ref.parts:
        if part not in outputs_dict.keys():
            break
        node = outputs_dict[part]
        outputs_dict = node
        if len(list(node.keys())) == 1 and '__name__' in node.keys():
            return node['__name__']

    return None

def raise_invalid_reference(ref, obj, name):
    """Takes the ref attempting to be looked-up,
    an obj that was the last model traversed too,
    the name that failed the next look-up.
    """
    raise InvalidPacoReference("""
Reference: {}

Reference look-up failed at '{}' trying to find name '{}'.
""".format(ref.raw, obj.paco_ref_parts, name))

def resolve_ref(ref_str, project, account_ctx=None, ref=None):
    """Resolve a reference"""
    if ref == None:
        ref = Reference(ref_str)
    ref_value = None
    if ref.type == "resource":
        try:
            ref_value = project['resource'][ref.parts[1]].resolve_ref(ref)
        except KeyError:
            raise_invalid_reference(ref, project['resource'], ref.parts[1])
    elif ref.type == "service":
        return get_resolve_ref_obj(project, project, ref, part_idx_start=0)
    elif ref.type == "netenv":
        try:
            obj = project['netenv']
            for i in range(1,4):
                obj = obj[ref.parts[i]]
        except KeyError:
            raise_invalid_reference(ref, obj, ref.parts[i])
        ref_value = get_resolve_ref_obj(project, obj, ref, 4)

    elif ref.type == "accounts":
        ref_value = resolve_accounts_ref(ref, project)
    elif ref.type == "function":
        ref_value = resolve_function_ref(ref, project, account_ctx)
    else:
        raise ValueError("Unsupported ref type: {}".format(ref.type))

    return ref_value

def function_ec2_ami_latest(ref, project, account_ctx):
    """EC2 AMI latest"""
    ami_description = None
    ami_name = None
    if ref.last_part == 'amazon-linux-2':
        ami_description = "Amazon Linux 2 AMI*"
        ami_name = 'amzn2-ami-hvm-*'
    elif ref.last_part == 'amazon-linux':
        ami_description = "Amazon Linux AMI*"
        ami_name = 'amzn-ami-hvm-*'
    elif ref.last_part == 'amazon-linux-nat':
        ami_description = "Amazon Linux AMI*"
        ami_name = 'amzn-ami-vpc-nat-hvm-*'
    elif ref.last_part == 'amazon-linux-2-ecs':
        ami_description = "Amazon Linux AMI*"
        ami_name = "amzn2-ami-ecs-hvm-*"
    else:
        raise ValueError("Unsupported AMI Name: {}".format(ref.last_part))

    cache_id = ami_name + ami_description
    if cache_id in ami_cache.keys():
        return ami_cache[cache_id]

    ec2_client = account_ctx.get_aws_client('ec2', aws_region=ref.region)
    filters = [ {
        'Name': 'name',
        'Values': [ami_name]
    },{
        'Name': 'description',
        'Values': [ami_description]
    },{
        'Name': 'architecture',
        'Values': ['x86_64']
    },{
        'Name': 'owner-alias',
        'Values': ['amazon']
    },{
        'Name': 'owner-id',
        'Values': ['*']
    },{
        'Name': 'state',
        'Values': ['available']
    },{
        'Name': 'root-device-type',
        'Values': ['ebs']
    },{
        'Name': 'virtualization-type',
        'Values': ['hvm']
    },{
        'Name': 'hypervisor',
        'Values': ['xen']
    },{
        'Name': 'image-type',
        'Values': ['machine']
    } ]

    ami_list= ec2_client.describe_images(
        Filters=filters,
        Owners=[
            'amazon'
        ]
    )
    # Sort on Creation date Desc
    image_details = sorted(ami_list['Images'],key=itemgetter('CreationDate'),reverse=True)
    ami_id = image_details[0]['ImageId']
    ami_cache[cache_id] = ami_id
    return ami_id

def import_and_call_function(ref, project, account_ctx):
    "Import a module and call a function in it with the Reference, Project and AccountContext"
    module_name = '.'.join(ref.parts[1:-1])
    if module_name.find(':') != -1:
        module_name = module_name.split(':')[0]
        function_name = module_name.split('.')[-1:][0]
        module_name = '.'.join(module_name.split('.')[:-1])
        module = importlib.import_module(module_name)
        extra_context = ref.raw.split(':')[1]
        return getattr(module, function_name)(ref, extra_context, project, account_ctx)
    else:
        function_name = ref.parts[-1:][0]
        module = importlib.import_module(module_name)
        return getattr(module, function_name)(ref, project, account_ctx)

def resolve_function_ref(ref, project, account_ctx):
    if account_ctx == None:
        return None
    if ref.ref.startswith('function.aws.ec2.ami.latest'):
        # ToDo: call this as a proper function ref? migrate it?
        return function_ec2_ami_latest(ref, project, account_ctx)
    else:
        return import_and_call_function(ref, project, account_ctx)

def resolve_accounts_ref(ref, project):
    "Return an IAccount object from the model."
    try:
        account = project[ref.parts[0]][ref.parts[1]]
    except KeyError:
        raise InvalidPacoReference("Can not resolve the reference '{}'".format(ref.raw))

    return account.resolve_ref(ref)




from paco.models import schemas
from paco.models.base import Resource, ApplicationResource, Named
from zope.interface import implementer
from zope.schema.fieldproperty import FieldProperty

@implementer(schemas.IEventTarget)
class EventTarget(Named):
    target = FieldProperty(schemas.IEventTarget['target'])
    input_json = FieldProperty(schemas.IEventTarget['input_json'])

@implementer(schemas.IEventsRule)
class EventsRule(ApplicationResource):
    description = FieldProperty(schemas.IEventsRule['description'])
    schedule_expression = FieldProperty(schemas.IEventsRule['schedule_expression'])
    enabled_state = FieldProperty(schemas.IEventsRule['enabled_state'])
    targets = FieldProperty(schemas.IEventsRule['targets'])

    def __init__(self, name, parent):
        super().__init__(name, parent)
        self.targets = []



from paco.models.locations import get_parent_by_interface
from paco.models import schemas

def get_formatted_model_context(obj):
    """Return a formatted string describing a model object in it's context
    """
    # ToDo: should work for all NetEnv objects, will need expanding to
    # handle Services and other objects
    try:
        netenv = get_parent_by_interface(obj, schemas.INetworkEnvironment)
        env = get_parent_by_interface(obj, schemas.IEnvironment)
        envreg = get_parent_by_interface(obj, schemas.IEnvironmentRegion)
        app = get_parent_by_interface(obj, schemas.IApplication)
        group = get_parent_by_interface(obj, schemas.IResourceGroup)
    except AttributeError:
        return 'Obj has no parent in model'
    # If the obj is only IParent, walk up to the first INamed for context
    named_obj = get_parent_by_interface(obj, schemas.INamed)
    out = ""
    if named_obj != None:
        out = "Paco reference: {}\n".format(named_obj.paco_ref)
    if netenv != None:
        if netenv.title:
            out += "Network Environment: {} ({})\n".format(netenv.title, netenv.name)
        else:
            out += "Network Environment: " + netenv.name + "\n"
    if env != None:
        if env.title:
            out += "Environment: {} ({})\n".format(env.title, env.name)
        else:
            out += "Environment: " + env.name + "\n"
    if envreg != None:
        out += "Account: " + envreg.network.aws_account + "\n"
        if envreg.title:
            out += "Region: {} ({})\n".format(envreg.title, envreg.name)
        else:
            out += "Region: " + envreg.name + "\n"
    if app != None:
        if app.title:
            out += "Application: {} ({})\n".format(app.title, app.name)
        else:
            out += "Application: " + app.name + "\n"
    if group != None:
        if group.title:
            out += "Resrouce Group: {} ({})\n".format(group.title, group.name)
        else:
            out += "Resource Group: " + group.name + "\n"
    return out

def smart_join(sep, list_of_strings):
    """
    Joins a list of strings and silently skips None entries
    """
    joined_string = ""
    first = True
    for item in list_of_strings:
        if item != None:
            if first == False:
                joined_string += sep
            joined_string += item
            first = False
    return joined_string


from paco.models import gen_vocabulary
from paco.models.exceptions import UnsupportedFeature


def ami_id(ref, date_and_type, project, account_ctx):
    date, ami_type = date_and_type.split('.')
    if date == 'latest':
        return gen_vocabulary.ami_ids[ref.region][ami_type][0]['ImageId']
    else:
        raise UnsupportedFeature("Only 'latest' is supported for AMI Id date")



from setuptools import setup

with open('README.md') as f:
    long_description = f.read()
    long_description += '\n\n'
with open('CHANGELOG.md') as f:
    long_description += f.read()

setup(
    name='paco.models',
    version='7.8.20',
    description='paco.models: Semantic cloud infrastructure configuration file format and object model',
    author='Waterbear Cloud',
    author_email='hello@waterbear.cloud',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/waterbear-cloud/paco.models',
    install_requires=['Setuptools', 'ruamel.yaml', 'zope.schema', 'troposphere'],
    packages=[
        'paco.models',
        'paco.aws',
    ],
    keywords=['AWS','Cloud','Infrastructure as Code'],
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        'License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)',
        'Development Status :: 5 - Production/Stable',
    ],
    include_package_data=True,
    zip_safe=False,
    package_dir={'': 'src'},
    entry_points={
        'console_scripts': [
            'paco_update_gen_vocabulary = paco.codegen:update_gen_vocabulary',
        ]
    },
)



from paco.models import schemas
from paco.models.base import Named
from paco.models.schemas import IProject, IVersionControl, IPacoWorkBucket, ISharedState
from zope.interface import implementer
from zope.schema.fieldproperty import FieldProperty
import paco.models.applications
import paco.models.networks
import paco.models.resources
import paco.models.metrics
import paco.models.logging


@implementer(IVersionControl)
class VersionControl(Named):
    def __init__(self, name, __parent__):
        super().__init__(name, __parent__)
        self.git_branch_environment_mappings = []

    enforce_branch_environments = FieldProperty(schemas.IVersionControl["enforce_branch_environments"])
    environment_branch_prefix = FieldProperty(schemas.IVersionControl["environment_branch_prefix"])
    git_branch_environment_mappings = FieldProperty(schemas.IVersionControl["git_branch_environment_mappings"])
    global_environment_name = FieldProperty(schemas.IVersionControl["global_environment_name"])

@implementer(IPacoWorkBucket)
class PacoWorkBucket(Named):
    def __init__(self, name, __parent__):
        super().__init__(name, __parent__)
        self.account = 'paco.ref accounts.master'

    bucket_name = FieldProperty(schemas.IPacoWorkBucket["bucket_name"])
    enabled = FieldProperty(schemas.IPacoWorkBucket["enabled"])
    account = FieldProperty(schemas.IPacoWorkBucket["account"])
    region = FieldProperty(schemas.IPacoWorkBucket["region"])

@implementer(ISharedState)
class SharedState(Named):
    def __init__(self, name, __parent__):
        super().__init__(name, __parent__)
        self.paco_work_bucket = PacoWorkBucket('paco_work_bucket', self)

    paco_work_bucket = FieldProperty(schemas.ISharedState["paco_work_bucket"])
    cloudformation_region = FieldProperty(schemas.ISharedState["cloudformation_region"])

@implementer(IProject)
class Project(Named, dict):
    """Paco project"""
    paco_project_version = FieldProperty(schemas.IProject["paco_project_version"])
    legacy_flags = FieldProperty(schemas.IProject["legacy_flags"])
    s3bucket_hash = FieldProperty(schemas.IProject["s3bucket_hash"])

    def __init__(self, name, __parent__):
        super().__init__(name, __parent__)
        self.legacy_flags = []
        self.resource_registry = {}

        # Version Control
        self.version_control = VersionControl('version_control', self)

        # Shared State
        self.shared_state = SharedState('shared_state', self)

        # Credentials
        self.credentials = paco.models.project.Credentials('credentials', self)
        self.credentials.title = 'Administrator Credentials'
        self.__setitem__('credentials', self.credentials)

        # Network Environments
        self.network_environments = paco.models.networks.NetworkEnvironments('netenv', self)
        self.network_environments.title = 'Network Environments'
        self.__setitem__('netenv', self.network_environments)

        # Services
        self.services = paco.models.services.Services('service', self)
        self.services.title = 'Services'
        self.__setitem__('service', self.services)

        # Accounts
        self.accounts = paco.models.accounts.Accounts('accounts', self)
        self.accounts.title = 'Cloud Accounts'
        self.__setitem__('accounts', self.accounts)

        # Global Resources
        self.resource = paco.models.resources.GlobalResources('resource', self)
        self.resource.title = 'Global Resources'
        self.__setitem__('resource', self.resource)

        # Init an empty S3Resource in-case there is no Resources/S3.yaml
        self.resource['s3'] = paco.models.resources.S3Resource('s3', self.resource)

        # IAM
        self.resource['iam'] = paco.models.resources.IAMResource('iam', self.resource)
        self.resource['iam'].title = 'IAM Resource'

        # SSM
        self.resource['ssm'] = paco.models.resources.SSMResource('ssm', self.resource)
        self.resource['ssm'].title = 'SSM Resource'

        # Monitoring
        self.monitor = paco.models.metrics.Monitor('monitor', self)

        # Alarm Sets
        self.monitor.alarm_sets = paco.models.metrics.AlarmSets('alarm_sets', self.monitor)

        # CloudWatch Logging
        self.monitor.cw_logging = paco.models.logging.CloudWatchLogging('cw_logging', self.resource)

    def get_all_resources_by_type(self, resource_type):
        """
        Every resource of a given type across ALL Applications.
        Should include NetEnv Apps and Service Apps or any Resource loaded in an Application
        """
        if resource_type not in self.resource_registry:
            return []
        return [
            resource for resource in self.resource_registry[resource_type].values()
        ]

    def find_object_from_cli(self, controller_type, component_name=None, config_name=None):
        found = None
        CONTROLLER_MAPPING = {
            'NetEnv' : 'netenv'
        }
        found = self[CONTROLLER_MAPPING[controller_type]]
        if component_name:
            found = found[component_name]
        if config_name:
            found = found[config_name]
        return found

    def get_all_applications(self):
        """
        Return a list of all applications in the model
        """
        results = []
        for ne in self['netenv'].values():
            for env in ne.values():
                for env_reg in env.env_regions.values():
                    for application in env_reg.applications.values():
                        results.append(application)
        return results

@implementer(schemas.ICredentials)
class Credentials(Named, dict):
    """
    Object attrs:
        - aws_access_key_id : string : AWS Access Key ID
        - aws_secret_access_key : string : AWS Secret Access Key
        - aws_default_region : string : AWS Default Region
        - master_account_id : string : Master AWS Account ID
        - master_admin_iam_username : string : Master Account Admin IAM Username
    """

    aws_access_key_id = FieldProperty(schemas.ICredentials["aws_access_key_id"])
    aws_secret_access_key = FieldProperty(schemas.ICredentials["aws_secret_access_key"])
    aws_default_region = FieldProperty(schemas.ICredentials["aws_default_region"])
    master_account_id = FieldProperty(schemas.ICredentials["master_account_id"])
    master_admin_iam_username = FieldProperty(schemas.ICredentials["master_admin_iam_username"])
    admin_iam_role_name = FieldProperty(schemas.ICredentials["admin_iam_role_name"])
    mfa_session_expiry_secs = FieldProperty(schemas.ICredentials["mfa_session_expiry_secs"])
    assume_role_session_expiry_secs = FieldProperty(schemas.ICredentials["assume_role_session_expiry_secs"])

    @property
    def mfa_role_arn(self):
        return 'arn:aws:iam::' + self.master_account_id + ':mfa/' + self.master_admin_iam_username

    def __init__(self, name, __parent__):
        super().__init__(name, __parent__)




#import paco.models.yaml
import ruamel.yaml
import troposphere
from ruamel.yaml.compat import StringIO
from paco.models.exceptions import TroposphereConversionError


yaml_str_tag = 'tag:yaml.org,2002:str'
yaml_seq_tag = 'tag:yaml.org,2002:seq'
yaml_map_tag = 'tag:yaml.org,2002:map'

def convert_yaml_node_to_troposphere(node):
    if node.tag == '!Sub':
        if type(node.value) == type(str()):
            # ScalarNode - single argument only
            return troposphere.Sub(node.value)
        else:
            values = {}
            for map_node in node.value[1:]:
                if map_node.tag != yaml_map_tag:
                    raise TroposphereConversionError("Substitue variables for !Sub must be mappings.")
                values[map_node.value[0][0].value] = map_node.value[0][1].value
            return troposphere.Sub(
                node.value[0].value,
                **values
            )

    elif node.tag == '!Ref':
        return troposphere.Ref(node.value)
    elif node.tag == '!Join':
        delimiter = node.value[0].value
        values = []
        for node in node.value[1].value:
            values.append(
                convert_yaml_node_to_troposphere(node)
            )
        return troposphere.Join(delimiter, values)

    elif node.tag == yaml_str_tag:
        return node.value
    else:
        raise TroposphereConversionError(
            "Unknown YAML to convert to Troposphere"
        )

def troposphere_constructor(loader, node):
    return convert_yaml_node_to_troposphere(node)

# def troposphere_representer(loader, node):
#     return export_troposphere_to_yaml_node(node)

cfn_tags = [
    '!Ref',
    '!Sub',
    '!Join',
    # '!Base64',
    # '!Cidr',
    # '!FindInMap',
    # '!GetAZs',
    # '!ImportValue',
    # '!Select',
    # '!Split',
    # '!Transform',
]

# ToDo: implement Fn::<Func> style syntax
# from ruamel.yaml.nodes import MappingNode
# from ruamel.yaml.constructor import BaseConstructor

# def construct_mapping(self, node, deep=False):
#     """SafeConstructor mapping that also handles Troposphere keys
#     such as Fn::Join: Fn::Sub: by creating Troposphere objects.
#     """
#     if isinstance(node, MappingNode):
#         self.flatten_mapping(node)
#     return BaseConstructor.construct_mapping(self, node, deep=deep)

# ruamel.yaml.SafeConstructor.construct_mapping = construct_mapping

class ModelYAML(ruamel.yaml.YAML):

    def add_troposphere_constructors(self):
        self.existing_constructors = {}
        for tag, constructor in self.constructor.yaml_constructors.items():
            if tag != None and tag.startswith('!'):
                self.existing_constructors[tag] = constructor
        for cfn_tag in cfn_tags:
            ruamel.yaml.add_constructor(
                cfn_tag,
                troposphere_constructor,
                constructor=ruamel.yaml.SafeConstructor
            )

    def restore_existing_constructors(self):
        for tag, constructor in self.existing_constructors.items():
            self.constructor.yaml_constructors[tag] = constructor

    def dump(self, data, stream=None, **kw):
        dumps = False
        if stream is None:
            dumps = True
            stream = StringIO()
        ruamel.yaml.YAML.dump(self, data, stream, **kw)
        if dumps:
            return stream.getvalue()

